use std::{
    env, fs,
    io::{self, Write},
    path::PathBuf,
    process,
    sync::OnceLock,
    thread,
};

use filesorters::Config;

static CONFIG: OnceLock<Config> = OnceLock::new();

fn main() {
    let config_path = get_config_path();

    if !config_path.exists() {
        println!("Config file does not exist, creating...");
        filesorters::Config::create(&config_path).unwrap();
    }

    match filesorters::Config::parse(config_path) {
        Ok(parsed_config) => {
            CONFIG.set(parsed_config).unwrap();
            println!("Config file loaded successfully");
        }
        Err(err) => {
            eprintln!("Error: {}", err);
            process::exit(1);
        }
    }
    {
        let mut counter = 1;
        for alias in CONFIG.get().unwrap().sources.keys() {
            println!("{}: {}", counter, alias);
            counter += 1;
        }
    }

    let mut actual_selection: Vec<String> = Vec::new();

    loop {
        print!("Enter: ");
        io::stdout().flush().unwrap();
        let mut selected = String::new();
        io::stdin()
            .read_line(&mut selected)
            .expect("Failed to read line");

        let parsed_selection: Vec<usize> = selected
            .split(" ")
            .filter(|y| !y.is_empty())
            .map(|y| y.replace("\n", ""))
            .flat_map(|x| x.parse::<usize>())
            .collect();

        if parsed_selection.is_empty() {
            println!("Invalid selection");
            continue;
        }
        if parsed_selection.len() > CONFIG.get().unwrap().sources.keys().len() {
            println!("Too many selections made");
            continue;
        }
        if parsed_selection
            .iter()
            .any(|x| *x > CONFIG.get().unwrap().sources.keys().len())
        {
            println!("Invalid selection");
            continue;
        }

        for selection in parsed_selection {
            actual_selection.push(
                CONFIG
                    .get()
                    .unwrap()
                    .sources
                    .keys()
                    .nth(selection - 1)
                    .unwrap()
                    .to_string(),
            );
        }

        break;
    }

    println!("Selected: {}", actual_selection.join(", "));

    let mut thread_pool: Vec<thread::JoinHandle<()>> = Vec::new();

    for selection in actual_selection {
        // TODO: forget the assumption that the user won't try to sort like 100 dirs at the same
        // time
        let thread = thread::spawn(move || {
            let _ = sort_files(selection, false); // TODO: handle errors
        });

        thread_pool.push(thread);
    }

    for thread in thread_pool {
        thread.join().unwrap();
    }
}

// TODO: implement recuriveness
fn sort_files(selection: String, _recursive: bool) -> Result<(), Box<dyn std::error::Error>> {
    println!("Sorting {}", selection);

    let search_path = CONFIG
        .get()
        .and_then(|config| config.sources.get(&selection))
        .cloned()
        .unwrap();

    println!("Searching in {}", search_path.display());

    let dir = fs::read_dir(search_path)?;

    let file_types = [
        (
            "Picture",
            filesorters::PICTURE_EXTENTIONS.to_vec(),
            CONFIG.get().unwrap().pictures_dir.clone(),
        ),
        (
            "Book",
            filesorters::BOOK_EXTENTIONS.to_vec(),
            CONFIG.get().unwrap().books_dir.clone(),
        ),
        (
            "Video",
            filesorters::VIDEO_EXTENTIONS.to_vec(),
            CONFIG.get().unwrap().videos_dir.clone(),
        ),
        (
            "Music",
            filesorters::SOUND_EXTENTIONS.to_vec(),
            CONFIG.get().unwrap().music_dir.clone(),
        ),
    ];

    for entry in dir.flatten().filter(|e| e.metadata().unwrap().is_file()) {
        if let Some(extension) = entry.path().extension().and_then(|e| e.to_str()) {
            let path = entry.path();

            for (label, valid_extensions, target_dir_option) in &file_types {
                if let Some(target_dir) = target_dir_option {
                    if valid_extensions.contains(&extension) {
                        println!("{}: {}", label, path.display());
                        move_file_to_directory(&path, target_dir);
                    }
                }
            }
        }
    }

    Ok(())
}

fn move_file_to_directory(path: &std::path::Path, dir: &std::path::Path) {
    let _ = fs::rename(path, dir.join(path.file_name().unwrap()));
}

fn get_config_path() -> PathBuf {
    if cfg!(target_os = "windows") {
        let username = env::var("USERNAME").unwrap_or_else(|_| "Default".to_string());
        PathBuf::from(format!(
            "C:\\Users\\{}\\AppData\\Local\\filesorters\\filesorters.toml",
            username
        ))
    } else {
        let xdg_config_home = env::var("XDG_CONFIG_HOME").ok();
        let home = env::var("HOME").ok();
        let user = env::var("USER").unwrap_or_else(|_| "default".to_string());

        xdg_config_home
            .map(|x| PathBuf::from(format!("{}/filesorters.toml", x)))
            .or_else(|| home.map(|h| PathBuf::from(format!("{}/.config/filesorters.toml", h))))
            .unwrap_or_else(|| PathBuf::from(format!("/home/{}/.config/filesorters.toml", user)))
    }
}
